@@ -7,7 +7,7 @@ __version__=''' $Id$ '''
 import sys, os, glob, random, unittest, traceback
 import preppy
 isPy310 = preppy.isPy310
-from check_basics import fposto
+from check_basics import fposto, mkSuite
 
 def checkErrorTextContains(texts,func,*args,**kwds):
     label = kwds.pop('label','')
@@ -366,10 +366,22 @@ catch all errors{{endtry}}"""
         self.assertEqual(self.getRunTimeOutput('{{import token as x}}{{x.__name__}}',quoteFunc=preppy.uStdQuote),'token')
 
     def checkFrom1(self):
-        self.assertEqual(self.getRunTimeOutput('{{from distutils import cmd}}{{cmd.__name__}}',quoteFunc=preppy.uStdQuote),'distutils.cmd')
+        if isPy310:
+            txt = '{{from tkinter import ttk}}{{ttk.__name__}}'
+            ans = 'tkinter.ttk'
+        else:
+            txt = '{{from distutils import cmd}}{{cmd.__name__}}'
+            ans = 'distutils.cmd'
+        self.assertEqual(self.getRunTimeOutput(txt,quoteFunc=preppy.uStdQuote),ans)
 
     def checkFrom2(self):
-        self.assertEqual(self.getRunTimeOutput('{{from distutils import cmd as x}}{{x.__name__}}',quoteFunc=preppy.uStdQuote),'distutils.cmd')
+        if isPy310:
+            txt = '{{from tkinter import ttk as x}}{{x.__name__}}'
+            ans = 'tkinter.ttk'
+        else:
+            txt = '{{from distutils import cmd as x}}{{x.__name__}}'
+            ans = 'distutils.cmd'
+        self.assertEqual(self.getRunTimeOutput(txt,quoteFunc=preppy.uStdQuote),ans)
 
     def checkAssert1(self):
         self.assertRaises(AssertionError,self.getRunTimeOutput,"{{assert i==1}}{{i}}", i=0, quoteFunc=preppy.uStdQuote)
@@ -782,12 +794,13 @@ class IncludeTestCase(PreppyTestCase):
                 )
 
 def makeSuite():
-    suite1 = unittest.makeSuite(NewGeneratedCodeTestCase,'check')
-    suite2 = unittest.makeSuite(GeneratedCodeTestCase,'check')
-    suite3 = unittest.makeSuite(OutputModeTestCase,'check')
-    suite4 = unittest.makeSuite(IncludeTestCase)
-    suite5 = unittest.makeSuite(ErrorIndicationTestCase,'check')
-    return unittest.TestSuite((suite1,suite2, suite3, suite4, suite5))
+    return unittest.TestSuite((
+                mkSuite(NewGeneratedCodeTestCase,'check'),
+                mkSuite(GeneratedCodeTestCase,'check'),
+                mkSuite(OutputModeTestCase,'check'),
+                mkSuite(IncludeTestCase),
+                mkSuite(ErrorIndicationTestCase,'check'),
+                ))
 
 if __name__=='__main__':
     runner = unittest.TextTestRunner()
