@@ -4,12 +4,13 @@
 # is on the path.
 import os, sys
 import unittest
+from check_basics import mkSuite
 
 def makeSuite():
     import check_basics, check_algorithms, check_load, check_lowlevel, test_import_hook
     suite = check_algorithms.makeSuite()
     suite.addTests(check_basics.suite)
-    suite.addTests(unittest.makeSuite(check_load.LoadTestCase,'load'))
+    suite.addTests(mkSuite(check_load.LoadTestCase,'load'))
     suite.addTests(check_lowlevel.makeSuite())
     suite.addTests(test_import_hook.makeSuite())
     return suite
